@@ -10,14 +10,13 @@ use crate::crypto::cipher::{AeadKey, Iv};
 use crate::crypto::tls13::{Hkdf, HkdfExpander, OkmBlock};
 use crate::enums::AlertDescription;
 use crate::error::Error;
+use crate::tls13::Tls13CipherSuite;
 use crate::tls13::key_schedule::{
     hkdf_expand_label, hkdf_expand_label_aead_key, hkdf_expand_label_block,
 };
-use crate::tls13::Tls13CipherSuite;
 
 #[cfg(feature = "std")]
 mod connection {
-    use alloc::sync::Arc;
     use alloc::vec;
     use alloc::vec::Vec;
     use core::fmt::{self, Debug};
@@ -27,13 +26,15 @@ mod connection {
 
     use super::{DirectionalKeys, KeyChange, Version};
     use crate::client::{ClientConfig, ClientConnectionData};
-    use crate::common_state::{CommonState, Protocol, DEFAULT_BUFFER_LIMIT};
+    use crate::common_state::{CommonState, DEFAULT_BUFFER_LIMIT, Protocol};
     use crate::conn::{ConnectionCore, SideData};
-    use crate::enums::{AlertDescription, ProtocolVersion};
+    use crate::enums::{AlertDescription, ContentType, ProtocolVersion};
     use crate::error::Error;
-    use crate::msgs::deframer::DeframerVecBuffer;
+    use crate::msgs::deframer::buffers::{DeframerVecBuffer, Locator};
     use crate::msgs::handshake::{ClientExtension, ServerExtension};
+    use crate::msgs::message::InboundPlainMessage;
     use crate::server::{ServerConfig, ServerConnectionData};
+    use crate::sync::Arc;
     use crate::vecbuf::ChunkVecBuffer;
 
     /// A QUIC client or server connection.
@@ -162,6 +163,23 @@ mod connection {
             quic_version: Version,
             name: ServerName<'static>,
             params: Vec<u8>,
+        ) -> Result<Self, Error> {
+            Self::new_with_alpn(
+                config.clone(),
+                quic_version,
+                name,
+                params,
+                config.alpn_protocols.clone(),
+            )
+        }
+
+        /// Make a new QUIC ClientConnection with custom ALPN protocols.
+        pub fn new_with_alpn(
+            config: Arc<ClientConfig>,
+            quic_version: Version,
+            name: ServerName<'static>,
+            params: Vec<u8>,
+            alpn_protocols: Vec<Vec<u8>>,
         ) -> Result<Self, Error> {
             if !config.supports_version(ProtocolVersion::TLSv1_3) {
                 return Err(Error::General(
@@ -180,7 +198,13 @@ mod connection {
                 Version::V1 | Version::V2 => ClientExtension::TransportParameters(params),
             };
 
-            let mut inner = ConnectionCore::for_client(config, name, vec![ext], Protocol::Quic)?;
+            let mut inner = ConnectionCore::for_client(
+                config,
+                name,
+                alpn_protocols,
+                vec![ext],
+                Protocol::Quic,
+            )?;
             inner.common_state.quic.version = quic_version;
             Ok(Self {
                 inner: inner.into(),
@@ -195,6 +219,11 @@ mod connection {
         pub fn is_early_data_accepted(&self) -> bool {
             self.inner.core.is_early_data_accepted()
         }
+
+        /// Returns the number of TLS1.3 tickets that have been received.
+        pub fn tls13_tickets_received(&self) -> u32 {
+            self.inner.tls13_tickets_received
+        }
     }
 
     impl Deref for ClientConnection {
@@ -212,7 +241,7 @@ mod connection {
     }
 
     impl Debug for ClientConnection {
-        fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
+        fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
             f.debug_struct("quic::ClientConnection")
                 .finish()
         }
@@ -312,7 +341,7 @@ mod connection {
     }
 
     impl Debug for ServerConnection {
-        fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
+        fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
             f.debug_struct("quic::ServerConnection")
                 .finish()
         }
@@ -371,13 +400,25 @@ mod connection {
         ///
         /// Handshake data obtained from separate encryption levels should be supplied in separate calls.
         pub fn read_hs(&mut self, plaintext: &[u8]) -> Result<(), Error> {
-            self.core.message_deframer.push(
-                ProtocolVersion::TLSv1_3,
-                plaintext,
-                &mut self.deframer_buffer,
-            )?;
+            let range = self.deframer_buffer.extend(plaintext);
+
+            self.core.hs_deframer.input_message(
+                InboundPlainMessage {
+                    typ: ContentType::Handshake,
+                    version: ProtocolVersion::TLSv1_3,
+                    payload: &self.deframer_buffer.filled()[range.clone()],
+                },
+                &Locator::new(self.deframer_buffer.filled()),
+                range.end,
+            );
+
+            self.core
+                .hs_deframer
+                .coalesce(self.deframer_buffer.filled_mut())?;
+
             self.core
                 .process_new_packets(&mut self.deframer_buffer, &mut self.sendable_plaintext)?;
+
             Ok(())
         }
 
@@ -680,7 +721,7 @@ pub trait PacketKey: Send + Sync {
     /// the additional authenticated data; and the `payload`. The authentication tag is returned if
     /// encryption succeeds.
     ///
-    /// Fails iff the payload is longer than allowed by the cipher suite's AEAD algorithm.
+    /// Fails if and only if the payload is longer than allowed by the cipher suite's AEAD algorithm.
     fn encrypt_in_place(
         &self,
         packet_number: u64,
@@ -722,9 +763,9 @@ pub trait PacketKey: Send + Sync {
     /// Once a `MessageDecrypter` produced for this suite has failed to decrypt `integrity_limit`
     /// messages, an attacker gains an advantage in forging messages.
     ///
-    /// This is not relevant for TLS over TCP (which is implemented in this crate)
-    /// because a single failed decryption is fatal to the connection.  However,
-    /// this quantity is used by QUIC.
+    /// This is not relevant for TLS over TCP (which is also implemented in this crate)
+    /// because a single failed decryption is fatal to the connection.
+    /// However, this quantity is used by QUIC.
     fn integrity_limit(&self) -> u64;
 }
 
@@ -797,6 +838,7 @@ impl<'a> KeyBuilder<'a> {
 }
 
 /// Produces QUIC initial keys from a TLS 1.3 ciphersuite and a QUIC key generation algorithm.
+#[derive(Clone, Copy)]
 pub struct Suite {
     /// The TLS 1.3 ciphersuite used to derive keys.
     pub suite: &'static Tls13CipherSuite,
@@ -874,7 +916,6 @@ impl Keys {
 /// Once the 1-RTT keys have been exchanged, either side may initiate a key update. Progressive
 /// update keys can be obtained from the [`Secrets`] returned in [`KeyChange::OneRtt`]. Note that
 /// only packet keys are updated by key updates; header protection keys remain the same.
-#[allow(clippy::large_enum_variant)]
 pub enum KeyChange {
     /// Keys for the handshake space
     Handshake {
